#uses "fwConfigs/fwPeriphAddressSNMP.ctl"
/**@name LIBRARY: unConfigSNMP.ctl

@author: Marcin Bes (BE/ICS)

Creation Date: 30/06/2017

Purpose:
This library contains the function used to import the SNMP front end.

Usage: Public

PVSS manager usage: Ctrl, NV, NG

Constraints:
  . unGeneration.cat
  . UNICOS-JCOP framework
  . PVSS version: 3.0
  . operating system: NT and W2000, but tested only under W2000.
  . distributed system: yes.
*/

// constant for SNMP system integrity
// const string c_unSystemIntegrity_SNMP = "SNMP_";
// constant for import of SNMP
const string UNSNMP_DPTYPE = "SNMP";
const string SNMP_POLL_PREFIX = "_SNMP_POLL_";

const string c_unSystemIntegrity_UNSNMP = "SNMP_";

const int c_UNSNMP_SECURITYLEVEL_NOAUTHNOPRIV = 1;
const int c_UNSNMP_SECURITYLEVEL_AUTHNOPRIV = 2;
const int c_UNSNMP_SECURITYLEVEL_AUTHPRIV = 3;

const int c_UNSNMP_AUTHPROTOCOL_NONE = 1;
const int c_UNSNMP_AUTHPROTOCOL_MD5 = 2;
const int c_UNSNMP_AUTHPROTOCOL_SHA = 3;

const int c_UNSNMP_PRIVPROTOCOL_NONE = 1;
const int c_UNSNMP_PRIVPROTOCOL_DES = 2;
const int c_UNSNMP_PRIVPROTOCOL_AES128 = 4;
const int c_UNSNMP_PRIVPROTOCOL_AES192 = 20;
const int c_UNSNMP_PRIVPROTOCOL_AES256 = 21;

/*importation line example: 
PLCCONFIG;SNMP;SNMP2;SNMP2_APP;138.137.12.14;1.3.6.1.8.2.1.1.1$visiblestring$0;1000;V3;162;2000;2;public;public;;;;1;1;;1;;spare
*/
const int UN_CONFIG_SNMP_LENGTH = 21;

const int UN_CONFIG_SNMP_AGENT_IP = 4;
const int UN_CONFIG_SNMP_COUNTER = 5;
const int UN_CONFIG_SNMP_POLL_INTERVAL = 6;
const int UN_CONFIG_SNMP_AGENT_VERSION = 7; // should be 1,2 or 3
const int UN_CONFIG_SNMP_AGENT_PORT = 8; // any string
const int UN_CONFIG_SNMP_AGENT_TIMEOUT = 9; // any string
const int UN_CONFIG_SNMP_AGENT_RETRIES = 10; // any string
const int UN_CONFIG_SNMP_AGENT_READCOMMUNITY = 11; // any string
const int UN_CONFIG_SNMP_AGENT_WRITECOMMUNITY = 12; // any string
const int UN_CONFIG_SNMP_AGENT_CONTEXTNAME = 13; // any string
const int UN_CONFIG_SNMP_AGENT_CONTEXTENGINEID = 14; // any string
const int UN_CONFIG_SNMP_AGENT_SECURITYNAME = 15; // any string
const int UN_CONFIG_SNMP_AGENT_SECURITYLEVEL = 16; // any string
const int UN_CONFIG_SNMP_AGENT_AUTHPROTOCOL = 17; // any string
const int UN_CONFIG_SNMP_AGENT_AUTHPASSWORD = 18; // any string
const int UN_CONFIG_SNMP_AGENT_PRIVPROTOCOL = 19; // any string
const int UN_CONFIG_SNMP_AGENT_PRIVPASSWORD = 20; // any string
const int UN_CONFIG_SNMP_AGENT_ADDITIONAL = 21; // any string

/** unConfigSNMP_checkDeleteCommand
Purpose: check the coherency of the delete command

Parameters:
  dsConfig, dyn_string, input, config dyn_string
  dsDeleteDps, dyn_string, output, list of dp that will be deleted
  exceptionInfo, dyn_string, output, for errors

*/
unConfigSNMP_checkDeleteCommand(dyn_string dsConfig, dyn_string &dsDeleteDps, dyn_string &exceptionInfo) {
    string sFeName, sPLCSubApplication = "", sObjectType = "", sNumber = "", sDpPlcName, sPlcModPlc;
    
    if (dynlen(dsConfig) >= 2) {			// At least, delete instruction and plc name
        // get PLC Name
        sFeName = dsConfig[2];
        if ((sFeName != "") && (dsConfig[1] == UN_DELETE_COMMAND)) {
            if (dynlen(dsConfig) >= 3) { // get Application
                sPLCSubApplication = dsConfig[3];
            }
            if (dynlen(dsConfig) >= 4) { // get DeviceType
                sObjectType = dsConfig[4];
            }
            if (dynlen(dsConfig) >= 5) { // get device number
                sNumber = dsConfig[5];
            }
            unGenericDpFunctions_getPlcDevices("", sFeName, sPLCSubApplication, sObjectType, sNumber, dsDeleteDps);
            // if only delete key word then delete also modbus and systemalarm
            if ((sFeName != "") && (sPLCSubApplication == "") && (sObjectType == "") && (sNumber == "")) {
                // sDpPlc is SNMP_plcName
                sDpPlcName = c_unSystemIntegrity_UNSNMP + sFeName;
                if (dpExists(sDpPlcName)) {
                    dyn_string dsAlarmDps;
                    dynAppend(dsDeleteDps, sDpPlcName);
                    dsAlarmDps = dpNames(c_unSystemAlarm_dpPattern + "*" + substr(sDpPlcName, strpos(sDpPlcName, ":") + 1), c_unSystemAlarm_dpType);
                    dynAppend(dsDeleteDps, dsAlarmDps);
                }
            }
        } else {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkDeleteCommand:" + getCatStr("unGeneration", "BADINPUTS"), "");
        }
    } else {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkDeleteCommand:" + getCatStr("unGeneration", "BADINPUTS"), "");
    }
}

/** unConfigSNMP_check
Purpose: check the coherency of the front-end config

Parameters:
  currentObject: string, input, type of config
  currentLineSplit: dyn_string, input, config
  dsDeleteDps: dyn_string, input, dp to be deleted
  exceptionInfo: dyn_string, output, for errors

*/
unConfigSNMP_check(string currentObject, dyn_string currentLineSplit, dyn_string dsDeleteDps, dyn_string &exceptionInfo) {
    switch (currentObject) {
        case UN_PLC_COMMAND:
            unConfigSNMP_checkApplication(currentLineSplit, dsDeleteDps, exceptionInfo);
            break;
        default:
            fwException_raise(exceptionInfo, "ERROR", getCatStr("unGeneration", "UNKNOWNFUNCTION"), "");
            break;
    }
}

/** unConfigSNMP_deleteCommand
Purpose: delete function

Parameters:
  dsConfig, dyn_string, input, config dyn_string
  exceptionInfo, dyn_string, output, for errors
*/
unConfigSNMP_deleteCommand(dyn_string dsConfig, dyn_string &exceptionInfo) {
    string sPLCSubApplication = "", sObjectType = "", sNumber = "";
    dyn_string dsDpeList, exceptionInfoTemp;

    unConfigSNMP_checkDeleteCommand(dsConfig, dsDpeList, exceptionInfoTemp);
    if (dynlen(exceptionInfoTemp) > 0) {
        dynAppend(exceptionInfo, exceptionInfoTemp);
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_deleteCommand:" + getCatStr("unGeneration", "BADINPUTS"), "");
    } else {
        if (dynlen(dsConfig) >= 2) {			// At least, delete instruction and plc name
            if (dynlen(dsConfig) >= 3) {
                sPLCSubApplication = dsConfig[3];
            }
            if (dynlen(dsConfig) >= 4) {
                sObjectType = dsConfig[4];
            }
            if (dynlen(dsConfig) >= 5) {
                sNumber = dsConfig[5];
            }
            unGenericDpFunctions_deletePlcDevices(dsConfig[2], sPLCSubApplication, sObjectType, sNumber, "progressBar", exceptionInfo);
            if ((dsConfig[2] != "") && (sPLCSubApplication == "") && (sObjectType == "") && (sNumber == "")) {
                unConfigSNMP_deleteFrontEnd(dsConfig[2], exceptionInfo);
                // remove all the SNMP polling groups
                dyn_string dsPollingGroups = unConfigSNMP_getAllPollingGroups(dsConfig[UN_CONFIG_PLC_NAME]);
                for(int i=1; i<=dynlen(dsPollingGroups);i++){
                    dpDelete(dsPollingGroups[i], "_PollGroup");
                }
            }
        }
    }
}

/**
Purpose: get list of all the polling groups of given front-end

Parameters:
  sFEName: string, input, name of the front-end
*/
dyn_string unConfigSNMP_getAllPollingGroups(string sFEName){
    string sPollGroup = SNMP_POLL_PREFIX + sFEName;
    dyn_string dsPollGroups = makeDynString(sPollGroup);
    dynAppend(dsPollGroups, dpNames(SNMP_POLL_PREFIX + sFEName + "_*", "_PollGroup"));
    return dsPollGroups;
}

/**
Purpose: set front-end config

Parameters:
  currentObject: string, input, type of config
  currentLineSplit: dyn_string, input, config
  exceptionInfo: dyn_string, output, for errors
*/
unConfigSNMP_set(string currentObject, dyn_string currentLineSplit, dyn_string &exceptionInfo) {
    switch (currentObject) {
        case UN_PLC_COMMAND:
            unConfigSNMP_setApplication(currentLineSplit, exceptionInfo);
            break;
        default:
            fwException_raise(exceptionInfo, "ERROR", getCatStr("unGeneration", "UNKNOWNFUNCTION"), "");
            break;
    }
}

/**
Purpose: check data for a sub application (not comm)
TODO: update this description once development done
Parameters:
		dsPLC: dyn_string, input, SNMP configuration line in file + additional parameters
			dsPLC must contains :
			1. string, "QUANTUM" or "PREMIUM"
			2. string, PLC name
			3. string, PLC subapplication
			4. int, PLC address = PLC number
			(additional parameters)
			5. driver number
			6. boolean archive name
			7. analog archive name
		dsDeleteDps, dyn_string, input, datapoints that will be deleted before importation
		exceptionInfo: dyn_string, output, for errors
*/
unConfigSNMP_checkApplication(dyn_string dsPLC, dyn_string dsDeleteDps, dyn_string &exceptionInfo) {
    dyn_string dsPLCLine, dsPLCAdditionalData;
    int i, iRes, iDriverNum;
    string sPLCDp, sHostname, sAliasDp, tempDp;

    if (dynlen(dsPLC) == (UN_CONFIG_SNMP_LENGTH + UN_CONFIG_PLC_ADDITIONAL_LENGTH)) {
        dsPLCLine = dsPLC;
        for (i = 1; i <= UN_CONFIG_PLC_ADDITIONAL_LENGTH; i++) {
            dynAppend(dsPLCAdditionalData, dsPLCLine[UN_CONFIG_SNMP_LENGTH + 1]);
            dynRemove(dsPLCLine, UN_CONFIG_SNMP_LENGTH + 1);
        }
        unConfigSNMP_checkData(dsPLCLine, exceptionInfo);
        unConfigSNMP_checkAdditionalData(dsPLCAdditionalData, exceptionInfo);

        // first check if the PLC name is correct:
        tempDp = dsPLCLine[UN_CONFIG_PLC_NAME];
        if (!unConfigGenericFunctions_nameCheck(tempDp)) {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkApplication: " + dsPLCLine[UN_CONFIG_PLC_NAME] + getCatStr("unGeneration", "BADPLCNAME"), "");
            return;
        }

        // check if the PLC Application Name is correct:
        tempDp = dsPLCLine[UN_CONFIG_PLC_SUBAPPLICATION];
        if (!unConfigGenericFunctions_nameCheck(tempDp)) {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkApplication: " + dsPLCLine[UN_CONFIG_PLC_SUBAPPLICATION] + getCatStr("unGeneration", "BADPLCAPPLICATIONNAME"), "");
            return;
        }

        // sPLCDp is UNSNMP_plcName
        sPLCDp = getSystemName() +  c_unSystemIntegrity_UNSNMP + dsPLCLine[UN_CONFIG_PLC_NAME];
        sAliasDp = unGenericDpFunctions_dpAliasToName(dsPLCLine[UN_CONFIG_PLC_NAME]);

        if (substr(sAliasDp, strlen(sAliasDp) - 1) == ".") {
            sAliasDp = substr(sAliasDp, 0, strlen(sAliasDp) - 1);
        }
        if ((sAliasDp != "") && (sAliasDp != sPLCDp)) {
            if (dynContains(dsDeleteDps, sAliasDp) <= 0) {	// Alias exists, is not used by the "SNMP" config datapoint and will not be deleted
                fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkApplication: " + dsPLCLine[UN_CONFIG_PLC_NAME] + getCatStr("unGeneration", "BADHOSTNAMEALIAS") + sAliasDp, "");
            }
        }
        if (dpExists(sPLCDp) && (dynContains(dsDeleteDps, sPLCDp) <= 0)) {
            sHostname = unGenericDpFunctions_getAlias(sPLCDp);
            iRes = dpGet(sPLCDp + ".communication.driver_num", iDriverNum);
            if (iRes >= 0) {
                if (iDriverNum != dsPLCAdditionalData[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM]) {
                    fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkApplication: " + getCatStr("unGeneration", "BADEXISTINGPLCDRIVER") + iDriverNum, "");
                }
                if (sHostname != dsPLCLine[UN_CONFIG_PLC_NAME]) {
                    fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkApplication: " + getCatStr("unGeneration", "BADEXISTINGPLCHOSTNAME"), "");
                }
            }
        }
        
        // check if SNMP manager exists
        string sManagerDP = getSystemName() + "_" + dsPLCAdditionalData[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM] + "_SNMPManager";
        if (!dpExists(sManagerDP)) {
            unImportDevice_report(0, "WARNING: SNMP Manager number " + dsPLCAdditionalData[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM] + " does not exist. It will be added to the project during importation.");
        } else {
            unImportDevice_report(0, "INFO: SNMP Manager number " + dsPLCAdditionalData[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM] + " exists in the project.");
        }
        
        int iAgentNumber = unConfigSNMP_findSNMPAgent(dsPLCLine[UN_CONFIG_SNMP_AGENT_IP], dsPLCAdditionalData[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM], dsPLCLine[UN_CONFIG_SNMP_AGENT_VERSION]);
        if(iAgentNumber == 0) {
            unImportDevice_report(0, "WARNING: SNMP " + dsPLCLine[UN_CONFIG_SNMP_AGENT_VERSION] + " Agent with IP address " + dsPLCLine[UN_CONFIG_SNMP_AGENT_IP] + " does not exist for manager number " + dsPLCAdditionalData[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM] + ". It will be added to the project during importation.");
        } else {
            unImportDevice_report(0, "INFO: SNMP " + dsPLCLine[UN_CONFIG_SNMP_AGENT_VERSION] + " Agent with IP address " + dsPLCLine[UN_CONFIG_SNMP_AGENT_IP] + " exists for manager number " + dsPLCAdditionalData[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM] + ".");
        }

    } else {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkApplication: " + getCatStr("unGeneration", "BADINPUTS") + ", wrong length of the importation line.", "");
    }
}

/**
Purpose: check PLC data for configuration

Parameters:
		dsPLC: dyn_string, input, PLC configuration line in file
			dsPLC must contains :
			1. string, "QUANTUM" or "PREMIUM"
			2. string, name
			3. string, subApplication
			4. int, PLC address = PLC number
		exceptionInfo, dyn_string, output, for errors
*/
unConfigSNMP_checkData(dyn_string dsPLC, dyn_string &exceptionInfo) {
    unsigned uPLCNumber;
    if (dynlen(dsPLC) != UN_CONFIG_SNMP_LENGTH) {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkData:" + getCatStr("unGeneration", "BADINPUTS") + ", wrong lengt of SNMP importation line.", "");
    } else {
        unConfigGenericFunctions_checkPLCType(dsPLC[UN_CONFIG_PLC_TYPE], exceptionInfo);

        if (strrtrim(dsPLC[UN_CONFIG_PLC_NAME]) == "") {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkData:" + getCatStr("unGeneration", "ERRORPLCNAME"), "");
        }

        if (strrtrim(dsPLC[UN_CONFIG_PLC_SUBAPPLICATION]) == "") {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkData:" + getCatStr("unGeneration", "ERRORSUBAPPLI"), "");
        }
        
        string sCounterAddress = dsPLC[UN_CONFIG_SNMP_COUNTER], sSubindex, sTransformation, sPollInterval;
        unConfigSNMP_splitAddress(sCounterAddress, sTransformation, sSubindex, sPollInterval);
        unConfigSNMP_checkAddress(sCounterAddress, exceptionInfo);
        if(dynlen(exceptionInfo) > 0) {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkData: Bad Counter address: " + dsPLC[UN_CONFIG_SNMP_COUNTER], "");
        }
        
        if (dsPLC[UN_CONFIG_SNMP_AGENT_VERSION] != "V1" && 
            dsPLC[UN_CONFIG_SNMP_AGENT_VERSION] != "V2" && 
            dsPLC[UN_CONFIG_SNMP_AGENT_VERSION] != "V1/V2" && 
            dsPLC[UN_CONFIG_SNMP_AGENT_VERSION] != "V3") {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkData: Unsupported SNMP version: " + dsPLC[UN_CONFIG_SNMP_AGENT_VERSION], "");
        }
                
        if(dsPLC[UN_CONFIG_SNMP_AGENT_SECURITYLEVEL] != c_UNSNMP_SECURITYLEVEL_NOAUTHNOPRIV &&
           dsPLC[UN_CONFIG_SNMP_AGENT_SECURITYLEVEL] != c_UNSNMP_SECURITYLEVEL_AUTHNOPRIV &&
           dsPLC[UN_CONFIG_SNMP_AGENT_SECURITYLEVEL] != c_UNSNMP_SECURITYLEVEL_AUTHPRIV) {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkData: Unsupported SNMP security level: " + dsPLC[UN_CONFIG_SNMP_AGENT_SECURITYLEVEL], "");
        }
        
        if(dsPLC[UN_CONFIG_SNMP_AGENT_AUTHPROTOCOL] != c_UNSNMP_AUTHPROTOCOL_NONE &&
           dsPLC[UN_CONFIG_SNMP_AGENT_AUTHPROTOCOL] != c_UNSNMP_AUTHPROTOCOL_MD5 &&
           dsPLC[UN_CONFIG_SNMP_AGENT_AUTHPROTOCOL] != c_UNSNMP_AUTHPROTOCOL_SHA) {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkData: Unsupported SNMP authentication protocol: " + dsPLC[UN_CONFIG_SNMP_AGENT_AUTHPROTOCOL], "");
        }
        
        if(dsPLC[UN_CONFIG_SNMP_AGENT_PRIVPROTOCOL] != c_UNSNMP_PRIVPROTOCOL_NONE &&
           dsPLC[UN_CONFIG_SNMP_AGENT_PRIVPROTOCOL] != c_UNSNMP_PRIVPROTOCOL_DES &&
           dsPLC[UN_CONFIG_SNMP_AGENT_PRIVPROTOCOL] != c_UNSNMP_PRIVPROTOCOL_AES128 &&
           dsPLC[UN_CONFIG_SNMP_AGENT_PRIVPROTOCOL] != c_UNSNMP_PRIVPROTOCOL_AES192 &&
           dsPLC[UN_CONFIG_SNMP_AGENT_PRIVPROTOCOL] != c_UNSNMP_PRIVPROTOCOL_AES256) {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkData: Unsupported SNMP priv protocol: " + dsPLC[UN_CONFIG_SNMP_AGENT_PRIVPROTOCOL], "");
        }
    }
}

/**
Purpose: check additional data for configuration

Parameters:
		dsPLC: dyn_string, input, PLC additional data
			dsPLC must contains :
			1. unsigned, driver number
			2. string, boolean archive
			3. string, analog archive
		exceptionInfo, dyn_string, output, for errors
*/
unConfigSNMP_checkAdditionalData(dyn_string dsPLC, dyn_string &exceptionInfo) {
    unsigned uPLCDriver;

    if (dynlen(dsPLC) != UN_CONFIG_PLC_ADDITIONAL_LENGTH) {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkAdditionalData:" + getCatStr("unGeneration", "BADINPUTS"), "");
    } else {
        if (dsPLC[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_BOOL] == "") {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkAdditionalData: bool archive" + getCatStr("unGeneration", "ERRORARCHIVEEMPTY"), "");
        }
        if (dsPLC[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_ANALOG] == "") {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkAdditionalData: analog archive" + getCatStr("unGeneration", "ERRORARCHIVEEMPTY"), "");
        }
        if (dsPLC[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_EVENT] == "") {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_checkAdditionalData: event archive" + getCatStr("unGeneration", "ERRORARCHIVEEMPTY"), "");
        }
    }
}

/**
Purpose: set PLC configuration for a sub application (not comm)

Parameters:
		dsPLC: dyn_string, input, PLC configuration line in file + additional parameters
			dsPLC must contains :
			(Plc config line)
			1. string, "QUANTUM" or "PREMIUM"
			2. string, PLC name
			3. string, PLC subapplication
			4. int, PLC address = PLC number
			(additional parameters)
			5. driver number
			6. boolean archive name
			7. analog archive name
		exceptionInfo: dyn_string, output, for errors
*/
unConfigSNMP_setApplication(dyn_string dsPLCData, dyn_string &exceptionInfo) {
    int iPLCNumber, iRes, i, position;
    string sDpName, addressReference, tempDp;
    dyn_string dsPLC, dsAdditionalParameters, dsSubApplications, dsImportTimes, dsBoolArchives, dsAnalogArchives, exceptionInfoTemp;
    dyn_string dsEventArchives;
    // Check
    
    if (dynlen(dsPLCData) != (UN_CONFIG_SNMP_LENGTH + UN_CONFIG_PLC_ADDITIONAL_LENGTH)) {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_setApplication: " + getCatStr("unGeneration", "BADINPUTS"), "");
        return;
    }

    // 1. Separate config file info and additional parameters
    dsPLC = dsPLCData;
    dynClear(dsAdditionalParameters);
    for (i = 1; i <= UN_CONFIG_PLC_ADDITIONAL_LENGTH; i++) {
        dynAppend(dsAdditionalParameters, dsPLC[UN_CONFIG_SNMP_LENGTH + 1]);
        dynRemove(dsPLC, UN_CONFIG_SNMP_LENGTH + 1);
    }

    // check if the PLC Application Name is correct:
    tempDp = dsPLCData[UN_CONFIG_PLC_SUBAPPLICATION];
    if (!unConfigGenericFunctions_nameCheck(tempDp)) {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_setApplication: " + dsPLCData[UN_CONFIG_PLC_SUBAPPLICATION] + getCatStr("unGeneration", "BADPLCAPPLICATIONNAME"), "");
        return;
    }

    int iDriverNumber = dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM];
    
    // add SNMP manager if don't exist (with the same number as WINCCsnmp manager)
    string managerDP = getSystemName() + "_" + iDriverNumber + "_SNMPManager";
    if (!dpExists(managerDP)) {
        DebugN("Creating SNMP manager " + managerDP);
        unConfigGenericFunctions_createDp(managerDP, "_SNMPManager", exceptionInfoTemp);
        if (dynlen(exceptionInfoTemp) > 0) {
            dynAppend(exceptionInfo, exceptionInfoTemp);
            return;
        }
    }
    
    // add SNMP agent if don't exist
    string sAgentIP = dsPLCData[UN_CONFIG_SNMP_AGENT_IP];
    string sFEName = dsPLC[UN_CONFIG_PLC_NAME];
    string sAgentVersion = dsPLC[UN_CONFIG_SNMP_AGENT_VERSION];
    // check if exists and if not, find first free number
    int iAgentNumber = unConfigSNMP_findSNMPAgent(sAgentIP, iDriverNumber, sAgentVersion);
    
    // no agent with given IP exists under this manager
    if(iAgentNumber == 0) {
        do {    // find first not existing agent
            iAgentNumber++;
            iRes = unConfigSNMP_findSNMPAgent("*", iDriverNumber, sAgentVersion, iAgentNumber);
        } while(iRes != 0);
        DebugN("Creating SNMP agent " + sAgentVersion + " number: " + iAgentNumber);
        unConfigSNMP_createAgent(iDriverNumber,
                                 iAgentNumber, 
                                 sFEName, 
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_IP], 
                                 exceptionInfoTemp,
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_VERSION],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_READCOMMUNITY],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_WRITECOMMUNITY],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_PORT],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_TIMEOUT],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_RETRIES],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_SECURITYNAME],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_SECURITYLEVEL],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_CONTEXTNAME],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_CONTEXTENGINEID],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_AUTHPROTOCOL],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_AUTHPASSWORD],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_PRIVPROTOCOL],
                                 dsPLCData[UN_CONFIG_SNMP_AGENT_PRIVPASSWORD]
                                 );
        if (dynlen(exceptionInfoTemp) > 0) {
            dynAppend(exceptionInfo, exceptionInfoTemp);
            return;
        }
    }
        
    // Set FronEnd configuration in UnSNMP
    sDpName = c_unSystemIntegrity_UNSNMP + sFEName;
    tempDp = sDpName;
    unConfigGenericFunctions_createDp(sDpName, UNSNMP_DPTYPE, exceptionInfoTemp);
    if (dynlen(exceptionInfoTemp) > 0) {
        dynAppend(exceptionInfo, exceptionInfoTemp);
        return;
    }

    sDpName = getSystemName() + sDpName;
    unConfigGenericFunctions_setAlias(sDpName, sFEName, exceptionInfo);
    iRes = dpGet(sDpName + ".configuration.subApplications", dsSubApplications,
                 sDpName + ".configuration.importTime", dsImportTimes,
                 sDpName + ".configuration.archive_bool", dsBoolArchives,
                 sDpName + ".configuration.archive_analog", dsAnalogArchives,
                 sDpName + ".configuration.archive_event", dsEventArchives);
    position = dynContains(dsSubApplications, dsPLC[UN_CONFIG_PLC_SUBAPPLICATION]);
    if (position <= 0) {
        position = dynlen(dsSubApplications) + 1;
    }
    // if the event archive is empty, set it to bool archive
    if (dynlen(dsEventArchives) <= 0) {
        dsEventArchives = dsBoolArchives;
    }

    dsSubApplications[position] = dsPLC[UN_CONFIG_PLC_SUBAPPLICATION];
    dsImportTimes[position] = (string)getCurrentTime();
    dsBoolArchives[position] = dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_BOOL];
    dsAnalogArchives[position] = dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_ANALOG];
    dsEventArchives[position] = dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_ARCHIVE_EVENT];
    iRes = dpSetWait(sDpName + ".configuration.importTime", dsImportTimes,
                     sDpName + ".configuration.archive_bool", dsBoolArchives,
                     sDpName + ".configuration.archive_analog", dsAnalogArchives,
                     sDpName + ".configuration.archive_event", dsEventArchives,
                     sDpName + ".configuration.subApplications", dsSubApplications,
                     sDpName + ".communication.driver_num", iDriverNumber,
                     sDpName + ".configuration.agent_number", iAgentNumber,
                     sDpName + ".configuration.agent_version", sAgentVersion);

    if (iRes < 0) {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_setApplication: " + getCatStr("unPVSS", "DPSETFAILED"), "");
    }

    // 4. Create _UnSystemAlarm datapoints
    unSystemIntegrity_createSystemAlarm(tempDp,
                                        DS_pattern,
                                        getCatStr("unSystemIntegrity", "PLC_DS_DESCRIPTION") + sFEName + " -> DS driver " + dsAdditionalParameters[UN_CONFIG_PLC_ADDITIONAL_DRIVER_NUM],
                                        exceptionInfo);

    // 4. Create polling group(s)
    dyn_string dsPollIntervals = strsplit(dsPLC[UN_CONFIG_SNMP_POLL_INTERVAL], ",");    // at least one

    string sPollGroup = SNMP_POLL_PREFIX + sFEName;
    dpCreate(sPollGroup, "_PollGroup");
    dpSetWait(sPollGroup + ".PollInterval", dsPollIntervals[1],
              sPollGroup + ".Active", true);

    for(int i=2; i<=dynlen(dsPollIntervals); i++){
        dpCreate(sPollGroup + "_" + i, "_PollGroup");
        dpSetWait(sPollGroup + "_" + i + ".PollInterval", dsPollIntervals[i],
                  sPollGroup + "_" + i + ".Active", true);
    }
    
    // set counter
    string sAgentType = strtolower(sAgentVersion)=="v3" ? "SNMPV3Entity" : "SNMPAgent";
    string sAgentDp = "_" + iDriverNumber + "_" + sAgentType + "_" +  iAgentNumber;
    if (!dpExists(sAgentDp)) {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_setApplication: dp " + sAgentDp + " does not exist.", "");
    }
    else {
        string sCounterAddress = dsPLC[UN_CONFIG_SNMP_COUNTER];
        string sSubindex="", sTransformation="", sPollInterval="";
        unConfigSNMP_splitAddress(sCounterAddress, sTransformation, sSubindex, sPollInterval);
        if(sSubindex == "") sSubindex = "0";
        string sPollGroup = unConfigSNMP_getPollGroupName(sFEName, sPollInterval);
        int iDataType = unConfigSNMP_dataTypeStringToInt(sTransformation);
        unConfigSNMP_setAddress(sDpName + ".communication.counter",
                                sFEName, 
                                sCounterAddress,
                                DPATTR_ADDR_MODE_INPUT_POLL, 
                                iDriverNumber, 
                                iDataType, 
                                sSubindex, 
                                iAgentNumber, 
                                sAgentVersion,
                                sPollGroup,
                                exceptionInfo);
    }
}

/**
Purpose: find a polling group for sFEName front-end with given sPollInterval

Parameters:
        sFEName: string, input, name of the front-end
        sPollInterval: string, input, value of the requested poll interval; empty value indicates that default polling group should be used
Returns:
        name of the polling group (default if not found)
*/
string unConfigSNMP_getPollGroupName(string sFEName, string sPollInterval){
    if(sPollInterval == ""){
        return SNMP_POLL_PREFIX + sFEName;
    } else {
        dyn_string dsPollingGroups = unConfigSNMP_getAllPollingGroups(sFEName);
        for(int i=1; i<=dynlen(dsPollingGroups); i++){
            int iPollInterval;
            dpGet(dsPollingGroups[i] + ".PollInterval", iPollInterval);
            if (iPollInterval == (int)sPollInterval){
                return dsPollingGroups[i];
            }
        }
    }
    unImportDevice_report(0, "WARNING: There is no Polling Group with interval " + sPollInterval + " for Front-End " + sFEName + ". Using default poll group.");
    return SNMP_POLL_PREFIX + sFEName;
}

/**
Purpose: create SNMP agent

Parameters:
        iManagerNumber: int, input, number of the SNMP manger
        iAgentNumber: int, input, number of the SNMP agent
        sFEName: string, input, FrontEnd name, used in description of the agent
        sIPAddress: string, input, IP address of the agent
        exceptionInfo: dyn_string, output, for errors
        sVersion: string, input, version of the agent; can be V1, V2, V1/V2, V3
        sReadCommunity: string, input, ReadCommunity agent parameter (V1 and V2 only)
        sWriteCommunity: string, input, WriteCommunity agent parameter (V1 and V2 only)
        iPort: int, input, port number, 161 by default
        iTimeout: int, input, timeout, unit is 1/100 second
        iRetries: int, input, number of retries (1 by defaul)
        sSecurityName: string, input, security name (V3 only)
        iSecurityLevel: int, input, security level 1=No authentication, 2=authentication, but no priv, 3=authentication + priv (V3 only)
        sContextName: string, input, context name (V3 only)
        sContextEngineID: string, input, context engine ID (V3 only)
        iAuthProtocol: int, input, authentication protocol 1=none, 2=MD5, 3=SHA (V3 only)
        sAuthPasswd: string, input, password for authorisation protocol (V3 only)
        iPrivProtocol: int, input, priv protocol, 1=none, 2=DES, 4=AES128, 20=AES192, 21=AES256 (V3 only)
        sPrivPasswd: string, input, password for priv protocol (V3 only)
*/
void unConfigSNMP_createAgent(int iManagerNumber, int iAgentNumber, string sFEName, string sIPAddress, dyn_string &exceptionInfo, string sVersion="V2", string sReadCommunity="public", string sWriteCommunity="public", int iPort=161, int iTimeout=100, int iRetries=1, string sSecurityName="", int iSecurityLevel=1, string sContextName="", string sContextEngineID="", int iAuthProtocol=1, string sAuthPasswd="", int iPrivProtocol=1, string sPrivPasswd=""){
    DebugN("Create agent "+ iAgentNumber + " (" + sFEName + ") in manager " + iManagerNumber + ", IP=" + sIPAddress + ", SNMP version " + sVersion + ", ReadCommunity=" + sReadCommunity + ", WriteCommunity="+sWriteCommunity + ", port =" + iPort + ", timeout=" + iTimeout + ", retries=" + iRetries);
    int iRes=0;
    string sAgentDPname;
    if(sVersion == "V1" || sVersion == "V2" || sVersion == "V1/V2"){
        DebugN("V1/V2 specific config: ReadCommunity=" + sReadCommunity + ", WriteCommunity="+sWriteCommunity);
        sAgentDPname = "_" + iManagerNumber + "_SNMPAgent_" + iAgentNumber;
        unConfigGenericFunctions_createDp(sAgentDPname, "_SNMPAgent", exceptionInfo);
        iRes = dpSetDescription(sAgentDPname + ".", "SNMPAgent_" + sFEName);
        iRes += dpSetWait(sAgentDPname + ".Access.IPAddress", sIPAddress,
            sAgentDPname + ".Access.ReadCommunity", sReadCommunity,
            sAgentDPname + ".Access.WriteCommunity", sWriteCommunity,
            sAgentDPname + ".Access.Port", iPort,
            sAgentDPname + ".Access.Retries", iRetries,
            sAgentDPname + ".Access.Timeout", iTimeout,
            sAgentDPname + ".Access.Protocol", (sVersion=="V1")?1:2
        );        
    } else if(sVersion == "V3") {
        DebugN("V3 specific config: SecurityName=" + sSecurityName + ", Security level: " + iSecurityLevel + ", Auth. Protocol = " + iAuthProtocol + ", Priv. Protocol=" + iPrivProtocol + ", ContextName=" + sContextName+ ", ContextEngineID=" + sContextEngineID);
        sAgentDPname = "_" + iManagerNumber + "_SNMPV3Entity_" + iAgentNumber;
        unConfigGenericFunctions_createDp(sAgentDPname, "_SNMPV3Entity", exceptionInfo);
        iRes = dpSetDescription(sAgentDPname + ".", "SNMPV3Entity_" + sFEName);
        iRes += dpSetWait(sAgentDPname + ".Access.IPAddress", sIPAddress,
                          sAgentDPname + ".Access.Port", iPort,
                          sAgentDPname + ".Access.Timeout", iTimeout,
                          sAgentDPname + ".Access.Retries", iRetries,
                          sAgentDPname + ".Access.SecurityName", sSecurityName,
                          sAgentDPname + ".Access.SecurityLevel", iSecurityLevel, // 1,2,3
                          sAgentDPname + ".Access.ContextName", sContextName,
                          sAgentDPname + ".Access.ContextEngineID", sContextEngineID,
                          sAgentDPname + ".Access.AuthProtocol", iAuthProtocol,  // 1,2,3
                          sAgentDPname + ".Access.AuthPasswd", sAuthPasswd,
                          sAgentDPname + ".Access.PrivProtocol", iPrivProtocol, // 1, 2, 4, 20, 21
                          sAgentDPname + ".Access.PrivPasswd", sPrivPasswd
        );
    } else {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_createAgent: SNMP Version " + iSNMPVersion + " is not supported.", "");
    }
    if (iRes < 0) {
        fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_createAgent: " + getCatStr("unPVSS", "DPSETFAILED"), "");
    }
}

/**
Purpose: split reference address into address, subindex and transformation type
if addressReference contains '$' separator, it becomes its first part
second part is saved in subindex

Parameters:
		sAddressReference, string, inout, value of the address
		sTransformation: string, out, transformation type name
        sSubindex: string, out, subindex
        sPollInterval: string, out, polling interval

*/
void unConfigSNMP_splitAddress(string &sAddressReference, string &sTransformation, string &sSubindex, string &sPollInterval){
    dyn_string dsAddressReference = strsplit(sAddressReference + " ", "$"); // adding " " fixes the problem with $ ending address reference
    int dsLen = dynlen(dsAddressReference);
    for(int i=1; i<=dsLen;i++){
        dsAddressReference[i] = strrtrim(dsAddressReference[i], " ");
    }
    if(dsLen >= 1) {
        sAddressReference = dsAddressReference[1];
        if(dsLen >= 2) {
            sTransformation = dsAddressReference[2];
            if(dsLen >= 3){
                sSubindex = dsAddressReference[3];
                if(dsLen >= 4){
                    sPollInterval = dsAddressReference[4];
                }
            }
        }
    }
    return;
}

/**
Purpose: check if SNMP agent with given IP address is existing for given manager number

Parameters:
		dsPLC: sIPAddress, input, IP address of agent; any IP if equal to *
		mixed: iManagerNumber, input, manager number (-num), mixed for compatibility with exiting fwAddress functions
        string sAgentVersion, input, version of the manager: V1, V2, V1/V2 or V3
        mixed sAgentNumber, inout, number of the agent, * if any

		exceptionInfo: dyn_string, output, for errors
*/
int unConfigSNMP_findSNMPAgent(string sIPAddress, mixed iManagerNumber, string sAgentVersion, mixed sAgentNumber = "*")
{
    DebugN("Search for " + sAgentVersion + " SNMP agent with IP:" + sIPAddress + " under manager number " + iManagerNumber + "; agentNumber = " + sAgentNumber);
    dyn_string agents;
    if(sAgentVersion == "V3"){
        agents = dpNames("_" + iManagerNumber + "_SNMPV3Entity_" + sAgentNumber, "_SNMPV3Entity"); // add V3: _SNMPV3Entity
    } else {
        agents = dpNames("_" + iManagerNumber + "_SNMPAgent_" + sAgentNumber, "_SNMPAgent"); // add V1/V2: _SNMPAgent
    }
    
    // search for agent with given IP address
    for(int i=1; i<=dynlen(agents); i++){
        string agentDP = agents[i];
        string sIP;
        dpGet(agentDP + ".Access.IPAddress", sIP);
        if(sIPAddress=="*" || sIP == sIPAddress) {
            dyn_string dsSplitName = strsplit(agentDP, "_");
            int iAgentNumber;
            sscanf(dsSplitName[dynlen(dsSplitName)], "%d", iAgentNumber);
            if(sAgentVersion == "V1" || sAgentVersion == "V2" ){
                int iSNMPVersion;
                dpGet(agentDP + ".Access.Protocol", iSNMPVersion);
                if(iSNMPVersion == 1 && sAgentVersion == "V1" || iSNMPVersion == 2 && sAgentVersion == "V2"){
                    return iAgentNumber;
                }
            } 
            else {
                return iAgentNumber;
            }
        }
    }
    
    // not found
    return 0;
    // fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_findAgentByIP: " + getCatStr("unPVSS", "DPSETFAILED"), "");    
}

/** 
Function checks whether addressReference follows the SNMP address format.

Parameters:
		dsPLC: sAddressReference, input, SNMP address
		exceptionInfo: dyn_string, output, for errors
*/

unConfigSNMP_checkAddress(string sAddressReference, dyn_string &exceptionInfo)
{
    if (sAddressReference != "") {
        if(strpos(sAddressReference, "1.3.6.1") != 0){
            fwException_raise(exceptionInfo,"ERROR"," fwPeriphAddressSNMP_check: Invalid OID \"" + sAddressReference + "\". OIDs must always begin with \"1.3.6.1\"","");
        }
    }
}

/** 
Function sets SNMP address for given dpe

Parameters:
        sDpeName: string, input, DataPoint name
        sFrontEndName: string, input, Front-End name
        sAddressReference: string, input, Address Value to be set
        iMode: int, input, address mode (in/out/inout)
        driverNum: mixed, input, Driver Number
        iDefaultConvertionDataType: int, input, data type
        sSubindex: string, input, subindex for the address
        iAgentID: int, input, ID of the agent
        sAgentVersion: int, input, version of the agent (V1, V2, V3)
        exceptionInfo: dyn_string, output, for errors
*/
unConfigSNMP_setAddress(string sDpeName, string sFrontEndName, string sAddressReference, int iMode, mixed driverNum, int iDefaultConvertionDataType, string sSubindex, int iAgentID, string sAgentVersion, string sPollGroup, dyn_string& exceptionInfo)
{
    if (sAddressReference == "") {
        fwPeriphAddress_delete(sDpeName, exceptionInfo);
    } else {
        
        if(sAgentVersion == "V1" || sAgentVersion == "V2") {
            sAgentVersion = "V1/V2";
        }
        
        dyn_string configParameters = makeDynString();
    
        configParameters[FW_PARAMETER_FIELD_COMMUNICATION] = fwPeriphAddress_TYPE_SNMP;
        configParameters[FW_PARAMETER_FIELD_DRIVER] = driverNum;
        configParameters[FW_PARAMETER_FIELD_ADDRESS] = sAddressReference;
        configParameters[FW_PARAMETER_FIELD_MODE] = iMode;
        configParameters[FW_PARAMETER_FIELD_DATATYPE] = iDefaultConvertionDataType;
        configParameters[FW_PARAMETER_FIELD_ACTIVE] = true;
        configParameters[fwPeriphAddress_SNMP_POLL_GROUP] = sPollGroup;
        configParameters[FW_PARAMETER_FIELD_SUBINDEX] = sSubindex;
        configParameters[fwPeriphAddress_SNMP_AGENT_ID] = iAgentID;
        configParameters[fwPeriphAddress_SNMP_AGENT_VERSION] = sAgentVersion;
    
        fwPeriphAddress_set(sDpeName, configParameters, exceptionInfo);
        if (dynlen(exceptionInfo) > 0) {
            DebugTN(exceptionInfo);
        }
    }
}

/**
Purpose: retrieve address reference for given Data Point

Parameters :
	sDeviceDpeName, string, input, device dp
	sDeviceType, string, input, plc device type, not used, for compatibility
*/
string DRV_SNMP_convertToUnicosAddress(string sDeviceDpeName, string sDeviceType)
{
    mixed addr_ref;
    int configType, subindex, datatype;
    string pollgroup;
    dpGet(sDeviceDpeName + ":_address.._type", configType,
          sDeviceDpeName + ":_address.._reference" , addr_ref,
          sDeviceDpeName + ":_address.._datatype" , datatype,
          sDeviceDpeName + ":_address.._subindex" , subindex,
          sDeviceDpeName + ":_address.._poll_group", pollgroup);
    if (configType == DPCONFIG_NONE) { // if addr is not defined - return an empty string
        return "";
    }
    
    // address reference starts with {agentNumber}_ prefix -> remove it
    dyn_string dsAddresSplit = strsplit(addr_ref, "_");
    if(dynlen(dsAddresSplit)>1){
        addr_ref = dsAddresSplit[2];
    }
    int iPollInterval;
    dpGet(pollgroup + ".PollInterval", iPollInterval);
    
    // get the polling time from the front-end
    
    string result = addr_ref + "$" + unConfigSNMP_dataTypeIntToString(datatype) + "$" + subindex + "$" + iPollInterval;
    
    return result;
}

/**
Purpose: delete Front-end config

Parameters :
	sPlcName, string, input, plc name
	exceptionInfo, dyn_string, output, for errors
*/
unConfigSNMP_deleteFrontEnd(string sPlcName, dyn_string& exceptionInfo) {
    string sDpPlcName, sPlcModPlc = "", exceptionText;
    int i, length, iRes;
    dyn_string dsAlarmDps;


    sDpPlcName = c_unSystemIntegrity_UNSNMP + sPlcName;
    if (!dpExists(sDpPlcName)) {
        sDpPlcName = "";
    }

    if (sDpPlcName != "") {
        if (unGenericDpFunctions_getSystemName(sDpPlcName) == getSystemName()) {
            string driver;
            dpGet(sDpPlcName + ".communication.driver_num", driver);

            // 2. Delete _Un_Plc
            iRes = dpDelete(substr(sDpPlcName, strpos(sDpPlcName, ":") + 1));
            if (iRes < 0) {
                fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_deleteFrontEnd: " + substr(sDpPlcName, strpos(sDpPlcName, ":") + 1) + getCatStr("unGeneral", "UNOBJECTNOTDELETED"), "");
            }
            // 3. Delete alarm system
            dsAlarmDps = dpNames(c_unSystemAlarm_dpPattern + "*" + substr(sDpPlcName, strpos(sDpPlcName, ":") + 1), c_unSystemAlarm_dpType);
            length = dynlen(dsAlarmDps);
            for (i = 1; i <= length; i++) {
                _unSystemIntegrity_modifyApplicationAlertList(dsAlarmDps[i], false, exceptionInfo);
                iRes = dpDelete(dsAlarmDps[i]);
                if (iRes < 0) {
                    fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_deleteFrontEnd: " + dsAlarmDps[i] + getCatStr("unGeneral", "UNOBJECTNOTDELETED"), "");
                }
            }
            // TODO: 
            // delete SNMP agent? (by its dpe comment = FEName?)
            // delete SNMP server (if no agents under anymore?)
        } else {
            fwException_raise(exceptionInfo, "ERROR", "unConfigSNMP_deleteFrontEnd: " + sDpPlcName + getCatStr("unGeneral", "UNOBJECTNOTDELETEDBADSYSTEM"), "");
        }
    }
}

/**
Purpose: return the list of Dps that wil be archived during the set of the front-end

Parameters:
		sCommand: string, input, the current entry in the import file
		sDp: string, input, the given dp: PLC or SystemAlarm  Dp
		sFEType: string, input, the front-end type
		dsArchiveDp: dyn_string, output, list of archivedDp
*/
SNMP_getFrontEndArchiveDp(string sCommand, string sDp, string sFEType, dyn_string &dsArchiveDp) {
    switch (sCommand) {
        case UN_PLC_COMMAND:
            dsArchiveDp = makeDynString();
            break;
        default:
            dsArchiveDp = makeDynString();
            break;
    }
}

/**
Purpose: export SNM FE

Parameters:
		dsParam: dyn_string, input, parameters
		exceptionInfo: dyn_string, output, for errors
*/
SNMP_Com_ExportConfig(dyn_string dsParam, dyn_string &exceptionInfo) {
    dyn_string dsDpParameters;
    dyn_string dsBoolArchive, dsAnalogArchive, dsEventArchive;
    string sPlcName, sPlcDrv;
    dyn_string dsTemp;

    if (dynlen(dsParam) == UN_CONFIG_EXPORT_LENGTH) {
        sPlcName = dsParam[UN_CONFIG_EXPORT_PLCNAME];
        dpGet(c_unSystemIntegrity_UNSNMP + sPlcName + ".configuration.archive_bool", dsBoolArchive,
              c_unSystemIntegrity_UNSNMP + sPlcName + ".configuration.archive_analog", dsAnalogArchive,
              c_unSystemIntegrity_UNSNMP + sPlcName + ".configuration.archive_event", dsEventArchive,
              c_unSystemIntegrity_UNSNMP + sPlcName + ".communication.driver_num", sPlcDrv);

        // Delete
        dynClear(dsDpParameters);
        dynAppend(dsDpParameters, "#" + UN_DELETE_COMMAND);
        dynAppend(dsDpParameters, dsParam[UN_CONFIG_EXPORT_PLCNAME]);
        dynAppend(dsDpParameters, dsParam[UN_CONFIG_EXPORT_APPLICATION]);
        unExportDevice_writeDsStringToFile(dsDpParameters, UN_PARAMETER_DELIMITER);

        // PLCCONFIG
        dynClear(dsDpParameters);
        dynAppend(dsDpParameters, UN_PLC_COMMAND);
        dynAppend(dsDpParameters, "SNMP");
        dynAppend(dsDpParameters, dsParam[UN_CONFIG_EXPORT_PLCNAME]);
        dynAppend(dsDpParameters, dsParam[UN_CONFIG_EXPORT_APPLICATION]);

        // Add SNMP specific data
        
        // driver number = manager number
        int driver_num;
        dpGet(c_unSystemIntegrity_UNSNMP + sPlcName + ".communication.driver_num", driver_num);
        
        string agent_version;
        int agent_number;
        dpGet(c_unSystemIntegrity_UNSNMP + sPlcName + ".configuration.agent_version", agent_version,
              c_unSystemIntegrity_UNSNMP + sPlcName + ".configuration.agent_number", agent_number);
              
        string agentType = (strpos(agent_version, "V3") >= 0) ? "_SNMPV3Entity_" : "_SNMPAgent_";
        string snmpAgentDPName = "_" + driver_num + agentType + agent_number;
        
        string IPAddress;
        int port, timeout, retries;
        dpGet(snmpAgentDPName + ".Access.IPAddress", IPAddress,
              snmpAgentDPName + ".Access.Port", port, 
              snmpAgentDPName + ".Access.Timeout", timeout, 
              snmpAgentDPName + ".Access.Retries", retries);
              
        dynAppend(dsDpParameters, IPAddress);
        
        string counterAddr = DRV_SNMP_convertToUnicosAddress(c_unSystemIntegrity_UNSNMP + sPlcName + ".communication.counter", UNSNMP_DPTYPE);
        dynAppend(dsDpParameters, counterAddr);

        dyn_string dsPollGroups = unConfigSNMP_getAllPollingGroups(sPlcName);
        int iPollInterval;
        dpGet(dsPollGroups[1] + ".PollInterval", iPollInterval);
        string sPollInterval = iPollInterval;
        for(int i=2; i<=dynlen(dsPollGroups); i++){
            dpGet(dsPollGroups[i] + ".PollInterval", iPollInterval);
            sPollInterval = sPollInterval + "," + iPollInterval;
        }
        
        dynAppend(dsDpParameters, sPollInterval);
        dynAppend(dsDpParameters, agent_version);
        dynAppend(dsDpParameters, port);
        dynAppend(dsDpParameters, timeout);
        dynAppend(dsDpParameters, retries);
        
        string readCommunity="", writeCommunity="", contextName="",contextEngineID="", securityName="", authpassword="", privpassword="";
        int securityLevel = 1, authprotocol = 1, privprotocol = 1;
        
        if(strpos(agent_version, "V3") == -1) {
            dpGet(snmpAgentDPName + ".Access.ReadCommunity", readCommunity,
                  snmpAgentDPName + ".Access.WriteCommunity", writeCommunity);
        } else {
            dpGet(snmpAgentDPName + ".Access.SecurityName", securityName,
                  snmpAgentDPName + ".Access.SecurityLevel", securityLevel,
                  snmpAgentDPName + ".Access.ContextName", contextName,
                  snmpAgentDPName + ".Access.ContextEngineID", contextEngineID,
                  snmpAgentDPName + ".Access.AuthProtocol", authprotocol,
                  snmpAgentDPName + ".Access.PrivProtocol", privprotocol);
        }
        
        dynAppend(dsDpParameters, readCommunity);
        dynAppend(dsDpParameters, writeCommunity);
        
        dynAppend(dsDpParameters, contextName);
        dynAppend(dsDpParameters, contextEngineID);
        dynAppend(dsDpParameters, securityName);
        dynAppend(dsDpParameters, securityLevel);
        dynAppend(dsDpParameters, authprotocol);
        dynAppend(dsDpParameters, authpassword);
        dynAppend(dsDpParameters, privprotocol);
        dynAppend(dsDpParameters, privpassword);
        dynAppend(dsDpParameters, ""); // additional data, spare
        
        unExportDevice_writeDsStringToFile(dsDpParameters, UN_PARAMETER_DELIMITER);
    } else {
        fwException_raise(exceptionInfo, "ERROR", "SNMP_Com_ExportConfig(): Wrong Parameters", "");
    }

}

/**
Purpose: return the driver config of SNMP front-end

Parameters:
        iDriverNumber: int, input, the driver number
        sName: string, output, the name to display
        sDriverName: string, output, the manager name
        iManType: int, output, the manager type
        sCommandLine: int, output, the command line to start the manager
*/
SNMP_getFrontEndManagerConfig(int iDriverNumber, string &sName, string &sDriverName, int &iManType, string &sCommandLine) {
  sName = "SNMP";
  sDriverName = "WCCOAsnmp";
  iManType = DRIVER_MAN;
  sCommandLine = "-num "+iDriverNumber;
}

/**
Purpose: return the number of SNMP data type; empty string default to visiblestring

Parameters:
        type: string, input, name of the datatype

Returns:
    number of datatype, -1 if not existing
*/
int unConfigSNMP_dataTypeStringToInt(string type) {
  mapping dataTypes;
  dataTypes[""] = 666;
  dataTypes["undefined"] = 660;
  dataTypes["integer32"] = 661;
  dataTypes["unsigned32"] =662;  
  dataTypes["unsigned64"] = 663;
  dataTypes["objectid"] = 664;
  dataTypes["octet"] = 665;
  dataTypes["visiblestring"] = 666;
  dataTypes["date time"] = 667;
  dataTypes["macaddress"] = 668;
  dataTypes["ipaddress"] = 669;
  dataTypes["bitstring"] = 670;
  dataTypes["visiblehexstring"] = 673;
  
  if(mappingHasKey(dataTypes, strtolower(type))){
    return dataTypes[type];
  }
  else {
      DebugN("Requested type: " + type);
    return -1;
  }
}

/**
Purpose: return the number of SNMP data type

Parameters:
        type: int, input, name of the datatype
        
Returns:
    name of the datatype
*/
string unConfigSNMP_dataTypeIntToString(int type)
{
  mapping dataTypes;
  dataTypes[660] = "undefined";
  dataTypes[661] = "integer32";
  dataTypes[662] = "unsigned32";
  dataTypes[663] = "unsigned64";
  dataTypes[664] = "objectid";
  dataTypes[665] = "octet";
  dataTypes[666] = "visiblestring";
  dataTypes[667] = "date time";
  dataTypes[668] = "macaddress";
  dataTypes[669] = "ipaddress";
  dataTypes[670] = "bitstring";
  dataTypes[673] = "visiblehexstring";
  
  if(mappingHasKey(dataTypes, type)){
    return dataTypes[type];
  }
  else {
    DebugN("Requested type: " + type);
    return "";
  }
}
